#!/usr/bin/env python

"""
Interpretation of vCalendar content.

Copyright (C) 2014, 2015 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from email.mime.text import MIMEText
from imiptools.dates import get_datetime, to_utc_datetime
from vCalendar import iterwrite, parse, ParseError, to_dict, to_node
import email.utils

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

class Object:

    "Access to calendar structures."

    def __init__(self, fragment):
        self.objtype, (self.details, self.attr) = fragment.items()[0]

    def get_items(self, name, all=True):
        return get_items(self.details, name, all)

    def get_item(self, name):
        return get_item(self.details, name)

    def get_value_map(self, name):
        return get_value_map(self.details, name)

    def get_values(self, name, all=True):
        return get_values(self.details, name, all)

    def get_value(self, name):
        return get_value(self.details, name)

    def get_utc_datetime(self, name):
        return get_utc_datetime(self.details, name)

    def to_node(self):
        return to_node({self.objtype : [(self.details, self.attr)]})

    def to_part(self, method):
        return to_part(method, [self.to_node()])

    # Direct access to the structure.

    def __getitem__(self, name):
        return self.details[name]

    def __setitem__(self, name, value):
        self.details[name] = value

    def __delitem__(self, name):
        del self.details[name]

# Construction and serialisation.

def make_calendar(nodes, method=None):

    """
    Return a complete calendar node wrapping the given 'nodes' and employing the
    given 'method', if indicated.
    """

    return ("VCALENDAR", {},
            (method and [("METHOD", {}, method)] or []) +
            [("VERSION", {}, "2.0")] +
            nodes
           )

def parse_object(f, encoding, objtype=None):

    """
    Parse the iTIP content from 'f' having the given 'encoding'. If 'objtype' is
    given, only objects of that type will be returned. Otherwise, the root of
    the content will be returned as a dictionary with a single key indicating
    the object type.

    Return None if the content was not readable or suitable.
    """

    try:
        try:
            doctype, attrs, elements = obj = parse(f, encoding=encoding)
            if objtype and doctype == objtype:
                return to_dict(obj)[objtype][0]
            elif not objtype:
                return to_dict(obj)
        finally:
            f.close()

    # NOTE: Handle parse errors properly.

    except (ParseError, ValueError):
        pass

    return None

def to_part(method, calendar):

    """
    Write using the given 'method', the 'calendar' details to a MIME
    text/calendar part.
    """

    encoding = "utf-8"
    out = StringIO()
    try:
        to_stream(out, make_calendar(calendar, method), encoding)
        part = MIMEText(out.getvalue(), "calendar", encoding)
        part.set_param("method", method)
        return part

    finally:
        out.close()

def to_stream(out, fragment, encoding="utf-8"):
    iterwrite(out, encoding=encoding).append(fragment)

# Structure access functions.

def get_fragments(d, name):

    """
    Return all fragments from 'd' with the given 'name'. Each fragment is thus
    suitable for using to initialise Object instances.
    """

    fragments = []
    if d.has_key(name):
        for value in d[name]:
            fragments.append(dict([(name, value)]))
    return fragments

def get_items(d, name, all=True):

    """
    Get all items from 'd' for the given 'name', returning single items if
    'all' is specified and set to a false value and if only one value is
    present for the name. Return None if no items are found for the name or if
    many items are found but 'all' is set to a false value.
    """

    if d.has_key(name):
        values = d[name]
        if all:
            return values
        elif len(values) == 1:
            return values[0]
        else:
            return None
    else:
        return None

def get_item(d, name):
    return get_items(d, name, False)

def get_value_map(d, name):

    """
    Return a dictionary for all items in 'd' having the given 'name'. The
    dictionary will map values for the name to any attributes or qualifiers
    that may have been present.
    """

    items = get_items(d, name)
    if items:
        return dict(items)
    else:
        return {}

def get_values(d, name, all=True):
    if d.has_key(name):
        values = d[name]
        if not all and len(values) == 1:
            return values[0][0]
        else:
            return map(lambda x: x[0], values)
    else:
        return None

def get_value(d, name):
    return get_values(d, name, False)

def get_utc_datetime(d, name):
    value, attr = get_item(d, name)
    dt = get_datetime(value, attr)
    return to_utc_datetime(dt)

def get_addresses(values):
    return [address for name, address in email.utils.getaddresses(values)]

def get_address(value):
    return value.lower().startswith("mailto:") and value.lower()[7:] or value

def get_uri(value):
    return value.lower().startswith("mailto:") and value.lower() or ":" in value and value or "mailto:%s" % value.lower()

def uri_dict(d):
    return dict([(get_uri(key), value) for key, value in d.items()])

def uri_item(item):
    return get_uri(item[0]), item[1]

def uri_items(items):
    return [(get_uri(value), attr) for value, attr in items]

# Operations on structure data.

def is_new_object(old_sequence, new_sequence, old_dtstamp, new_dtstamp, partstat_set):

    """
    Return for the given 'old_sequence' and 'new_sequence', 'old_dtstamp' and
    'new_dtstamp', and the 'partstat_set' indication, whether the object
    providing the new information is really newer than the object providing the
    old information.
    """

    have_sequence = old_sequence is not None and new_sequence is not None
    is_same_sequence = have_sequence and int(new_sequence) == int(old_sequence)

    have_dtstamp = old_dtstamp and new_dtstamp
    is_old_dtstamp = have_dtstamp and new_dtstamp < old_dtstamp or old_dtstamp and not new_dtstamp

    is_old_sequence = have_sequence and (
        int(new_sequence) < int(old_sequence) or
        is_same_sequence and is_old_dtstamp
        )

    return is_same_sequence and partstat_set or not is_old_sequence

# vim: tabstop=4 expandtab shiftwidth=4
